
function simultaneous_perturbation(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                                   state_variables::AbstractVector{OptimizationVariable},
                                   targets::Dict{Int, Target},
                                   comms::OptComms;
                                  ) where {S, T, E <: AbstractMagneticEquilibrium, D <: SPSA}
    rank = comms.compute.rank
    original_input = deepcopy(equilibrium_wrapper.input)
    eq = nothing
    θ = map(v -> v.value, state_variables)
    δθ = similar(θ)
    if comms.is_compute_head
        group_id = comms.inter_group.rank
        group_size = comms.inter_group.size
        target_values = Dict{Int, T}()
        finished, _ = MPI.recv(0, group_id, comms.inter_group.comm)

@debug "Group $(comms.inter_group.rank) has received $finished"

        MPI.bcast(finished, 0, comms.compute.comm)

        while !finished
            Δ, _ = MPI.recv(0, group_size + group_id, comms.inter_group.comm)
            op, _ = MPI.recv(0, 2 * group_size + group_id, comms.inter_group.comm)
            δθ .= op.(θ, Δ)
            MPI.bcast(δθ, 0, comms.compute.comm)
            foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
            translate!(equilibrium_wrapper, state_variables)
            eq = compute_equilibrium(equilibrium_wrapper, comms.compute.comm)
            MPI.send(group_id, 0, group_id, comms.inter_group.comm)

@debug "Group $group_id has sent index back to head"

            MPI.send(eq, 0, group_size + group_id, comms.inter_group.comm)

@debug "Group $group_id has sent eq back to head"

            MPI.send(op, 0, 2 * group_size + group_id, comms.inter_group.comm)

@debug "Group $group_id has sent $op back to head"

            if !isnothing(eq)
                update_equilibrium!(equilibrium_wrapper, eq)
                push!(target_values, compute_targets(targets, equilibrium_wrapper, comms.compute.comm)...)
                MPI.send(target_values, 0, 3 * group_size + group_id, comms.inter_group.comm)
            end
            set_input!(equilibrium_wrapper, original_input)
            finished, _ = MPI.recv(0, group_id, comms.inter_group.comm)

@debug "Group $(comms.inter_group.rank) has received $finished"

            MPI.bcast(finished, 0, comms.compute.comm)
        end
    else
        finished = MPI.bcast(nothing, 0, comms.compute.comm)
        if !finished            
            δθ = MPI.bcast(nothing, 0, comms.compute.comm)
            foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
            translate!(equilibrium_wrapper, state_variables)
            eq = compute_equilibrium(equilibrium_wrapper, comms.compute.comm)
            if !isnothing(eq)
                update_equilibrium!(equilibrium_wrapper, eq)
                compute_targets(targets, equilibrium_wrapper, comms.compute.comm)
            end
            set_input!(equilibrium_wrapper, original_input)
            finished = MPI.bcast(nothing, 0, comms.compute.comm)
        end
    end
end

function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target},
                  comms::OptComms;
                  iteration::Int = 1,
                  max_evals::Int = 10,
                  kwargs...
                ) where {T, E <: AbstractMagneticEquilibrium, D <: SPSA}

    if comms.is_world_head
        g = equilibrium_wrapper.grad_method
        group_size = comms.inter_group.size
        target_gradients = Dict{Int, Vector{T}}()
        forward_targets = empty(target_gradients)
        backward_targets = empty(target_gradients)
        gₖ = zeros(T, length(state_variables))
        Δ = similar(gₖ)
        cₖ = g.δ/(iteration^g.γ)
        group_status = fill(false, group_size-1)
        eq_status = fill(false, 2)
        eq_f = nothing
        eq_b = nothing
        eval_count = 0
        while (isnothing(eq_f) || isnothing(eq_b)) && eval_count < max_evals
            eq_f = nothing
            eq_b = nothing
            eq_status = fill(false, 2)
            b_dist = Bernoulli()
            Δ .= rand(b_dist, length(state_variables)) |> x -> cₖ * (2x .- 1)
            for group_id in eachindex(group_status)
                i = findfirst(j -> !j, eq_status) 
                if !isnothing(i)
                    op = if i == 1
                        +
                    else
                        -
                    end
                    MPI.send(false, group_id, group_id, comms.inter_group.comm)
                    MPI.send(Δ, group_id, group_size + group_id, comms.inter_group.comm)
                    MPI.send(op, group_id, 2 * group_size + group_id, comms.inter_group.comm)
                    group_status[group_id] = false
                    eq_status[i] = true
                else

                    MPI.send(true, group_id, group_id, comms.inter_group.comm)
                    group_status[group_id] = false
                end
            end

            eq_status = fill(false, 2)
            finished = false 
            while !finished
                index, _ = MPI.recv(-2, -1, comms.inter_group.comm)
                eq, _ = MPI.recv(index, group_size + index, comms.inter_group.comm)
                op, _ = MPI.recv(index, 2 * group_size + index, comms.inter_group.comm)
                if !isnothing(eq)
                    if op === +
                        forward_targets, _ = MPI.recv(index, 3 * group_size + index, comms.inter_group.comm)
                        eq_f = eq
                        eq_status[1] = true
                        group_status[index] = true
                    elseif op === -
                        backward_targets, _ = MPI.recv(index, 3 * group_size + index, comms.inter_group.comm)
                        eq_b = eq
                        eq_status[2] = true
                        group_status[index] = true
                    else
                        error("Fatal error in SPSA algorithm, recveived operation $op")
                    end
                else
                    if op === +
                        eq_f = eq
                        eq_status[1] = true
                        group_status[index] = true
                    elseif op === -
                        eq_b = eq
                        eq_status[2] = true
                        group_status[index] = true
                    else
                        error("Fatal error in SPSA algorithm, received operation $op")
                    end
                end
                finished = all(e -> e, eq_status)
            end
            eval_count += 1
        end
        for group_id in findall(g -> g, group_status)
            MPI.send(true, group_id, group_id, comms.inter_group.comm)
        end
        if !isnothing(eq_f) && !isnothing(eq_b)
            for key in keys(targets)
                forward_objective = (targets[key].target_value - forward_targets[key])^2/targets[key].target_weight^2
                backward_objective = (targets[key].target_value - backward_targets[key])^2/targets[key].target_weight^2
                target_gradients[key] = 0.5 * (forward_targets[key] - backward_targets[key]) ./ Δ
                gₖ .+= 0.5 * (forward_objective - backward_objective) ./ Δ
            end
            @debug "Final target_gradients: $target_gradients"
            aₖ = g.ϵ/((g.β + iteration)^g.α)
            @info "aₖ = $aₖ"
            @info "gₖ = $gₖ"
            @info "aₖ * gₖ = $(aₖ * gₖ)"
            return aₖ * gₖ, target_gradients
        else
            error("Fatal error in SPSA algorithm!")
        end
    else
        simultaneous_perturbation(equilibrium_wrapper, state_variables, targets, comms)
    end
end
